unit ExtHilit;

{  Electro-Concept Mauricie, 1999-2000 }
{ Implements TExtHighlighter object, used for scope dependant and prioritized syntax highlighting
  in a TPlusMemo }

{ Note: to install this component on your palette, add file ExtKeys in your package }

interface

uses Classes, Graphics, Controls, Plusmemo;
{$B-}  { not complete boolean evaluation }

{$IFDEF WIN32}
  {$H+}  { long strings }
  {$J+}  { writeable typed constants }
{$ENDIF}
{$IFNDEF DEBUG}
  {$R-}
  {$Q-}
{$ENDIF}


type
  TExtKeywordInfo = record  BasicInfo: TKeywordInfo; Scope, Priority: SmallInt end;
  TExtKeyInfoLen  = record BasicInfo: TKeyInfoLen; Scope, Priority: SmallInt end;
  pExtKeyInfoLen  = ^TExtKeyInfoLen;

  TExtStartStopInfo = record
                        BasicSSInfo: StartStopInfo;
                        Scope, Priority: SmallInt
                      end;

  pExStartStopInfo = ^TExtStartStopInfo;

  TExtKeywordList = class(TKeywordList)
    private
      procedure ReadData(Reader: TReader);
      procedure WriteData(Writer: TWriter);
      procedure SetKeyInfo(i: Integer; const ki: TExtKeywordInfo);
      function  GetKeyInfo(i: Integer): TExtKeywordInfo;
    protected
      procedure DefineProperties(Filer: TFiler); override;
    public
      destructor Destroy; override;
      function AddExtKeyWord(const KeyWord: string; Options: TWordOptions;
                             Style        : TExtFontStyles;
                             ContextNumber,
                             Scope,
                             Priority      : SmallInt;
                             Cursor        : TCursor;
                             Backgnd,
                             Foregnd       : TColor): Integer;

      procedure Assign(Source: TPersistent); override;
      procedure Clear; override;
      procedure Delete(i: Integer); override;
      property ExtKeyInfos[i: Integer]: TExtKeywordInfo read getKeyInfo write setKeyInfo;
    end;


  TExtStartStopList = class(TStartStopKeyList)
    private
      procedure ReadData(Reader: TReader);
      procedure WriteData(Writer: TWriter);
      function getExtStartStopInfo(I: Integer): TExtStartStopInfo;
      procedure setExtStartStopInfo(I: Integer; const ss: TExtStartStopInfo);
    protected
      procedure DefineProperties(Filer: TFiler); override;
    public
      constructor Create;
      function AddExStartStopKey(const StartKey, StopKey: string;
                                 Options         : TWordOptions;
                                 Style           : TExtFontStyles;
                                 ContextNumber,
                                 Scope, Priority : SmallInt;
                                 Cursor          : TCursor;
                                 Backgnd, Foregnd: TColor;
                                 EndAtPar        : Boolean): Integer;
      property ExtStartStopInfo[I: Integer]: TExtStartStopInfo read getExtStartStopInfo write setExtStartStopInfo;
    end;

  THighlightInfo = class(TPersistent)
    private
      fAltFont    : Boolean;
      fStyle      : TFontStyles;
      fBackground,
      fForeground : TColor;
      procedure setAltFont   (af: Boolean);
      procedure setStyle     (s: TFontStyles);
      procedure setBackground(bg: TColor);
      procedure setForeground(fg: TColor);
    protected
      fOnChange: TNotifyEvent;
    public
      procedure Assign(Source: TPersistent); override;
      property OnChange  : TNotifyEvent read fOnChange write fOnChange;
    published
      property AltFont   : Boolean     read fAltFont    write setAltFont;
      property Style     : TFontStyles read fStyle      write setStyle;
      property Background: TColor      read fBackground write setBackground;
      property Foreground: TColor      read fForeground write setForeground;
    end;


  TKeywordEvent = procedure (Sender: TObject; StartKey, StopKey: TPlusNavigator; KIndex: Integer; var Accept: Boolean) of object;
  TQueryHilitEvent = procedure (Sender: TObject; Scope: SmallInt; var DoHilit: Boolean) of object;

  TCustomExtHighlighter = class(TPlusHighlighter)
    private
      fKeywords      : TExtKeywordList;
      fStartStopKeys : TExtStartStopList;
      fDelimiters    : TSetOfChar;
      fSubHighlighter: TPlusHighlighter;
      fActive        : Boolean;

      fOnKeyword,
      fOnStop, fOnStart :TKeywordEvent;

      fAltFont       : Boolean;
      fContext       : SmallInt;
      fBackgnd       : TColor;
      fForegnd       : TColor;
      fStyle         : TFontStyles;
      fPriority      : Integer;
      fScope         : Integer;

      fOnQueryHilit  : TQueryHilitEvent;
      procedure SetSeparators(const sep: string);
      function  GetSeparators: string;
      procedure setAltFont(const Value: Boolean);
      procedure setBackgnd(const Value: TColor);
      procedure setForegnd(const Value: TColor);
      procedure setStyle(const Value: TFontStyles);
      procedure setPriority(const Value: Integer);
      procedure setScope(const Value: Integer);

    protected
      fNav1, fNav2, fNav3: TPlusNavigator;
      procedure Loaded; override;
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure ApplyKeywordsList(Start, Stop: TPlusNavigator; BaseIndex: Integer);  override;
      function  FindStart(Start, Stop: TPlusNavigator; BaseIndex: Integer): Boolean; override;
      function  FindStop(Start, Stop: TPlusNavigator; BaseIndex: Integer): Boolean; override;
      function  GetPriority(const d: DynInfoRec; BaseIndex: Integer): SmallInt;
      property Delimiters: TSetOfChar read fDelimiters write fDelimiters;

      { support for global highlighting from descendants }
      property AltFont: Boolean read fAltFont write setAltFont;
      property Background: TColor read fBackgnd write setBackgnd;
      property ContextNum: SmallInt read fContext write fContext;
      property Foreground: TColor read fForegnd write setForegnd;
      property Scope: Integer read fScope write setScope;
      property Priority: Integer read fPriority write setPriority;
      property Style: TFontStyles read fStyle write setStyle;
      property OnQueryHilit: TQueryHilitEvent read fOnQueryHilit write fOnQueryHilit;

      { support for extended keywords and start-stop key highlighting }
      property Keywords      : TExtKeywordList   read fKeywords      write fKeywords;
      property StartStopKeys : TExtStartStopList read fStartStopKeys write fStartStopKeys;
      property SubHighlighter: TPlusHighlighter  read fSubHighlighter write fSubHighlighter;
      property OnKeyword: TKeywordEvent read fOnKeyword write fOnKeyword;
      property OnStart  : TKeywordEvent read fOnStart   write fOnStart;
      property OnStop   : TKeywordEvent read fOnStop    write fOnStop;

    public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure  ReApplyKeys;
      property Separators: string read GetSeparators write SetSeparators;
      property Active: Boolean read fActive write fActive default True;
    end;

  TExtHighlighter = class(TCustomExtHighlighter)
    published
      property Active;
      property Keywords;
      property StartStopKeys;
      property Separators;
      property SubHighlighter;
      property OnKeyword;
      property OnStart;
      property OnStop;
    end;

const
  QueryPriority = 32767;
  EndAtDelScope = 32767;

implementation

uses SysUtils, Dialogs;

destructor TExtKeywordList.Destroy;
var i: Integer; pk: pExtKeyInfoLen;
begin
  {$IFDEF WIN32}
  { under D1, memory for keyword infos is freed
    in Delete even at destroy time, whereas with D2 and D3
    Delete is not called within Destroy }
  for i:= 0 to Count-1 do
    begin
      pk:= pExtKeyInfoLen(Objects[i]);
      if pk<>nil then Dispose(pk);
      Objects[i]:= nil
    end;
  {$ENDIF}
  inherited Destroy
end;

procedure TExtKeywordList.DefineProperties(Filer: TFiler);
begin
  Filer.DefineProperty('Strings', ReadData, WriteData, Count>0)
end;

procedure TExtKeywordList.Assign(Source: TPersistent);
var i: Integer; pk: pExtKeyInfoLen; isext: Boolean;
begin
  if Source is TKeywordList then
    begin
      isext:= Source is TExtKeywordList;
      BeginUpdate;
      Clear;
      for i:= 0 to TKeywordList(Source).Count-1 do
        begin
          Add(TKeywordList(Source)[i]);
          New(pk);
          if isext then pk^:= pExtKeyInfoLen(TKeywordList(Source).Objects[i])^
                   else
                     begin
                       pk^.BasicInfo:= pKeyInfoLen(TKeywordList(Source).Objects[i])^;
                       pk^.Scope:= 0;
                       pk^.Priority:= 0
                     end;
          Objects[i]:= TObject(pk)
        end;
      EndUpdate
    end
end;

procedure TExtKeywordList.Delete(i: Integer);
var pk: pExtKeyInfoLen;
begin
  pk:= pExtKeyInfoLen(Objects[i]);
  if pk<>nil then Dispose(pk);
  Objects[i]:= nil;
  inherited Delete(i)
end;

procedure TExtKeywordList.Clear;
var i: Integer; pk: pExtKeyInfoLen;
begin
  for i:= 0 to Count-1 do
    begin
      pk:= pExtKeyInfoLen(Objects[i]);
      if pk<>nil then Dispose(pk);
      Objects[i]:= nil
    end;
  inherited Clear
end;

procedure TExtKeywordList.ReadData(Reader: TReader);
var pk: pExtKeyInfoLen;
begin
  Reader.ReadListBegin;
  Clear;
  BeginUpdate;
  while not Reader.EndOfList do
    begin
      New(pk);
      AddObject(Reader.ReadString, TObject(pk));
      with pk^, BasicInfo.BasicPart do
        begin
          Byte(Options):= Reader.ReadInteger;
          Byte(Style)  := Reader.ReadInteger;
          ContextNumber:= Reader.ReadInteger;
          Cursor  := Reader.ReadInteger;
          Backgnd := Reader.ReadInteger;
          Foregnd := Reader.ReadInteger;
          Scope   := Reader.ReadInteger;
          Priority:= Reader.ReadInteger
        end
    end;
  Reader.ReadListEnd;
  EndUpdate
end;

procedure TExtKeywordList.WriteData(Writer: TWriter);
var i: Integer;
begin
  Writer.WriteListBegin;
  for i:= 0 to Count-1 do
    begin
      Writer.WriteString(Strings[i]);
      with pExtKeyInfoLen(Objects[i])^, BasicInfo.BasicPart do
        begin
          Writer.WriteInteger(Byte(Options));
          Writer.WriteInteger(Byte(Style));
          Writer.WriteInteger(ContextNumber);
          Writer.WriteInteger(Cursor);
          Writer.WriteInteger(Backgnd);
          Writer.WriteInteger(Foregnd);
          Writer.WriteInteger(Scope);
          Writer.WriteInteger(Priority)
        end
    end;
  Writer.WriteListEnd
end;

function TExtKeywordList.AddExtKeyWord(const KeyWord   : string;
                                       Options         : TWordOptions;
                                       Style           : TExtFontStyles;
                                       ContextNumber,
                                       Scope,
                                       Priority        : SmallInt;
                                       Cursor          : TCursor;
                                       Backgnd, Foregnd: TColor): Integer;
var dinfo: TExtKeywordInfo;
begin
  Result:= Add(KeyWord);
  dinfo.BasicInfo.Options:= Options;
  dinfo.BasicInfo.Style:= Style;
  dinfo.BasicInfo.ContextNumber:= ContextNumber;
  dinfo.BasicInfo.Cursor:= Cursor;
  dinfo.BasicInfo.Backgnd:= Backgnd;
  dinfo.BasicInfo.Foregnd:= Foregnd;
  dinfo.Scope:= Scope;
  dinfo.Priority:= Priority;
  ExtKeyInfos[Result]:= dinfo
end;

procedure TExtKeywordList.setKeyInfo(i: Integer; const ki: TExtKeywordInfo);
var pk: pExtKeyInfoLen;
begin
  pk:= pExtKeyInfoLen(Objects[i]);
  if pk=nil then New(pk);
  pk^.BasicInfo.BasicPart:= ki.BasicInfo;
  pk^.Priority:= ki.Priority;
  pk^.Scope:= ki.Scope;
  Objects[i]:= TObject(pk)
end;

function TExtKeywordList.getKeyInfo(i: Integer): TExtKeywordInfo;
var pk: pExtKeyInfoLen;
begin
  pk:= pExtKeyInfoLen(Objects[i]);
  if pk<>nil then
    begin
      Result.BasicInfo:= pk^.BasicInfo.BasicPart;
      Result.Scope:= pk^.Scope;
      Result.Priority:= pk^.Priority
    end
  else
    with Result.BasicInfo do
      begin
        Options:= [];
        Style:= [];
        ContextNumber:= 0;
        Backgnd:= 0;
        Foregnd:= 0;
        Cursor:= crDefault;
        Result.Scope:= 0;
        Result.Priority:= 0
      end
end;


constructor TExtStartStopList.Create;
begin
  inherited Create;
  fElementSize:= SizeOf(TExtStartStopInfo);
  fElementSpace:= ((fElementSize-1) div 16 +1) * 16;
end;

procedure TExtStartStopList.DefineProperties(Filer: TFiler);
begin
  Filer.DefineProperty('Keys', ReadData, WriteData, Count>0)
end;

procedure TExtStartStopList.ReadData(Reader: TReader);
var pk: TExtStartStopInfo; ps: Boolean;  start, stop: string;
begin
  Reader.ReadListBegin;
  Clear;
  while not Reader.EndOfList do
    begin
      start:= Reader.ReadString;
      stop:= Reader.ReadString;
      with pk, pk.BasicSSInfo.Attributes do
        begin
          Byte(Options):= Reader.ReadInteger;
          Byte(Style):= Reader.ReadInteger;
          ContextNumber:= Reader.ReadInteger;
          Cursor:= Reader.ReadInteger;
          Backgnd:= Reader.ReadInteger;
          Foregnd:= Reader.ReadInteger;
          Scope  := Reader.ReadInteger;
          Priority:= Reader.ReadInteger;
          ps:= Reader.ReadBoolean;
          AddExStartStopKey(start, stop, Options, Style, ContextNumber, Scope, Priority, Cursor,
                                   Backgnd, Foregnd, ps);
        end;
    end;
  Reader.ReadListEnd;
  fDelChecked:= False;
end;

procedure TExtStartStopList.WriteData(Writer: TWriter);
var i: Integer;
begin
  Writer.WriteListBegin;
  for i:= 0 to Count-1 do
    with pExStartStopInfo(Pointers[i])^ do
      with BasicSSInfo do
        begin
          Writer.WriteString(StrPas(StartKey));
          Writer.WriteString(StrPas(StopKey));
          Writer.WriteInteger(Byte(Attributes.Options));
          Writer.WriteInteger(Byte(Attributes.Style));
          Writer.WriteInteger(Attributes.ContextNumber);
          Writer.WriteInteger(Attributes.Cursor);
          Writer.WriteInteger(Attributes.Backgnd);
          Writer.WriteInteger(Attributes.Foregnd);
          Writer.WriteInteger(Scope);
          Writer.WriteInteger(Priority);
          Writer.WriteBoolean(ParStop)
        end;
  Writer.WriteListEnd
end;

function TExtStartStopList.AddExStartStopKey(const StartKey, StopKey: string;
                                             Options         : TWordOptions;
                                             Style           : TExtFontStyles;
                                             ContextNumber,
                                             Scope, Priority : SmallInt;
                                             Cursor          : TCursor;
                                             Backgnd, Foregnd: TColor;
                                             EndAtPar        : Boolean): Integer;
var pss : pExStartStopInfo;
begin
  Result:= AddStartStopKey(StartKey, StopKey, Options, Style, ContextNumber, Cursor, Backgnd, Foregnd, EndAtPar);
  pss:= Pointers[Result];
  pss^.Scope:= Scope;
  pss^.Priority:= Priority
end;

function TExtStartStopList.getExtStartStopInfo(I: Integer): TExtStartStopInfo;
begin
  Result:= pExStartStopInfo(Pointers[I])^
end;

procedure TExtStartStopList.setExtStartStopInfo(I: Integer; const ss: TExtStartStopInfo);
begin
  pExStartStopInfo(Pointers[I])^:= ss
end;

procedure THighlightInfo.Assign(Source: TPersistent);
var srch: THighlightInfo;
begin
  if Source is THighlightInfo then
    begin
      srch:= THighlightInfo(Source);
      fAltFont   := srch.fAltFont;
      fStyle     := srch.fStyle;
      fBackground:= srch.fBackground;
      fForeground:= srch.fForeground;
      if Assigned(fOnChange) then fOnChange(Self)
    end
  else inherited Assign(Source)
end;

procedure THighlightInfo.setAltFont(af: Boolean);
begin
  if af<>fAltFont then
    begin
      fAltFont:= af;
      if Assigned(fOnChange) then fOnChange(Self)
    end
end;

procedure THighlightInfo.setStyle(s: TFontStyles);
begin
  if s<>fStyle then
    begin
      fStyle:= s;
      if Assigned(fOnChange) then fOnChange(Self)
    end
end;

procedure THighlightInfo.setBackground(bg: TColor);
begin
  if bg<>fBackground then
    begin
      fBackground:= bg;
      if Assigned(fOnChange) then fOnChange(Self)
    end
end;

procedure THighlightInfo.setForeground(fg: TColor);
begin
  if fg<>fForeground then
    begin
      fForeground:= fg;
      if Assigned(fOnChange) then fOnChange(Self)
    end
end;

constructor TCustomExtHighlighter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fStartStopKeys:= TExtStartStopList.Create;
  fKeywords:= TExtKeywordList.Create;
  fBackgnd:= -1; fForegnd:= -1;
  fNav1:= TPlusNavigator.Create(nil);
  fNav2:= TPlusNavigator.Create(nil);
  fNav3:= TPlusNavigator.Create(nil);
  fDelimiters:= [#9, ' ', '.', ',', ';'];
  fActive:= True
end;

destructor TCustomExtHighlighter.Destroy;
begin
  fStartStopKeys.Free;
  fKeywords.Free;
  fNav1.Free;
  fNav2.Free;
  fNav3.Free;
  inherited Destroy
end;

procedure TCustomExtHighlighter.Loaded;
begin
  inherited Loaded;
  if csDesigning in ComponentState then ReApplyKeys
end;

procedure TCustomExtHighlighter.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if Byte(Operation)=pm_Msg then
    begin   { propagate context messages coming from AComponent }
      if fSubHighlighter<>nil then TCustomExtHighlighter(fSubHighlighter).Notification(AComponent, Operation)
    end
  else
    begin
      if (AComponent=fSubHighlighter) and (Operation=opRemove) then fSubHighlighter:= nil;
      inherited Notification(AComponent, Operation)
    end
end;

procedure TCustomExtHighlighter.ReApplyKeys;
var i: Integer;
begin
  for i:= 0 to MemoList.Count-1 do
      TPlusMemo(MemoList[i]).ReApplyKeywords
end;

procedure TCustomExtHighlighter.SetSeparators(const sep: string);
var i: Integer; del: TSetOfChar;
begin
  del:= [#0..#26];
  for i:= 1 to Length(sep) do del:= del + [sep[i]];
  if del<>fDelimiters then
    begin
      fDelimiters:= del;
      if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then ReApplyKeys
    end
end;

function TCustomExtHighlighter.GetSeparators: string;
var i: Char;
begin
  Result:= '';
  for i:= #27 to #255 do
      if i in fDelimiters then Result:= Result + i
end;

procedure TCustomExtHighlighter.setAltFont(const Value: Boolean);
begin
  if Value<>fAltFont then
    begin
      fAltFont := Value;
      if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then ReApplyKeys
    end
end;

procedure TCustomExtHighlighter.setBackgnd(const Value: TColor);
begin
  if Value<>fBackgnd then
    begin
      fBackgnd := Value;
      if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then ReApplyKeys
    end
end;

procedure TCustomExtHighlighter.setForegnd(const Value: TColor);
begin
  if Value<>fForegnd then
    begin
      fForegnd := Value;
      if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then ReApplyKeys
    end
end;

procedure TCustomExtHighlighter.setStyle(const Value: TFontStyles);
begin
  if Value<>fStyle then
    begin
      fStyle := Value;
      if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then ReApplyKeys
    end
end;

procedure TCustomExtHighlighter.setPriority(const Value: Integer);
begin
  if Value<>fPriority then
    begin
      fPriority := Value;
      if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then ReApplyKeys
    end
end;

procedure TCustomExtHighlighter.setScope(const Value: Integer);
begin
  if Value<>fScope then
    begin
      fScope := Value;
      if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then ReApplyKeys
    end
end;


function TCustomExtHighlighter.GetPriority(const d: DynInfoRec; BaseIndex: Integer): SmallInt;
var kindex: SmallInt;
begin
  Result:= 0;
  if (d.DynStyle and $80 <> 0) and (d.Level>=0) then
    begin
      kindex:= d.KeyIndex[d.Level];
      if kindex>=0 then Result:= High(Result)
      else
        begin
          kindex:= kindex and $7fff;
          if (kindex>=BaseIndex) and (kindex-BaseIndex<StartStopKeys.Count) then
                 Result:= pExStartStopInfo(StartStopKeys.Pointers[kindex-BaseIndex])^.Priority
        end
    end
end;

type TPPlusMemo = class(TPlusMemo);

function getlevel(const dyn: DynInfoRec): SmallInt;
begin
  if dyn.DynStyle and $c0 <>$c0 then Result:= -1
                                else Result:= dyn.Level
end;

function dyntoscope(const drec: DynInfoRec): Integer;
begin   { returns the scope }
  if drec.DynStyle and $80 = 0 then Result:= 0
                               else Result:= drec.Context
end;

procedure TCustomExtHighlighter.ApplyKeywordsList(Start, Stop: TPlusNavigator; BaseIndex: Integer);
  type keywfound = record keyw: Integer; Offset: PlusCardinal; keylen: PlusCardinal end;
  var i: Longint; j: Integer;
      tlow, tup, tfound, tsearch, s: PChar;
      { tlow, tup:  initial case text; tfound: working text; s: keyword we search }

      tlen, slen: PlusCardinal;
      dinfop: pExtKeyInfoLen;
      pp: pParInfo;
      scar: Char;
      startoff, stopoff: PlusCardinal;
      tstartoffset: PlusCardinal;  { offset of tlow, tup in par.Text }

      startp, stopp: Longint;

      nbkeywtodo: SmallInt;
      keywtodo: array[0..15] of keywfound;
      kwsome: Boolean;
      endword: PlusCardinal;
      smemo: TPPlusMemo;
      slowcar, supcar: Char;
      kcount: Integer;

  procedure applykwfound;
    var skey1, skey2: TPlusNavigator; needed: PlusCardinal; ok, inkeyword: Boolean;
        drec: DynInfoRec;  kinfo: pExtKeyInfoLen; k, m, currentlevel, currentscope, currentpriority: SmallInt;
    begin
    skey1:= TPlusNavigator.Create(Start.fPMemo);
    skey2:= TPlusNavigator.Create(Start.fPMemo);
    with skey1 do
      begin
      fPar:= pp; fPos:= pp^.StartOffset;
      fDynNb:= 0; fOffset:= 0;
      fParLine:= 0;
      fParNb:= i;
      end;
    with pp^ do
      begin
      needed:= DynCount+nbkeywtodo*2;
      if needed>DynBufLen then
        begin
        {$IFNDEF WIN32} pmReAllocMem(DynBuffer, DynBufLen*DCodeLen, needed*DCodeLen);
        {$ELSE}         ReAllocMem(DynBuffer, needed*DCodeLen); {$ENDIF}
        DynBufLen:= needed
        end
      end;

    for k:= 0 to nbkeywtodo-1 do
       with keywtodo[k] do
         begin
         skey1.Pos:= pp^.StartOffset+offset;
         skey1.RightOfDyn;
         currentlevel:= getlevel(skey1.pDynAttr^);
         currentscope:= DynToScope(skey1.pDynAttr^);
         currentpriority:= GetPriority(skey1.pDynAttr^, BaseIndex);
         TObject(kinfo):= Keywords.Objects[keyw];
         if (currentlevel<15) and
            ((kinfo^.Priority=QueryPriority) or
            (((kinfo^.Scope=0) and (kinfo^.Priority>currentpriority)) or (kinfo^.Scope=currentscope))) then
           begin
           skey2.Assign(skey1);
           skey2.Pos:= skey2.Pos+ keylen;
           ok:= True; inkeyword:= False;
           if skey2.DynNb>skey1.fDynNb then
             begin
             m:= skey1.fDynNb+1;
             while (m<=skey2.fDynNb) and ok do
               with skey1.fPar^.DynBuffer^[m] do
                 begin
                 if DynStyle and $c0<>$c0 then begin ok:= inkeyword; inkeyword:= False end
                 else
                   if Level<currentlevel then ok:= False
                     else
                       if KeyIndex[Level]<0 then ok:= False
                                            else inkeyword:= True;
                 Inc(m)
                 end;
             if inkeyword then ok:= False
             end;

           if ok and Assigned(fOnKeyword) then fOnKeyword(Self, skey1, skey2, keyw, ok);
           if ok then
             begin
             kwsome:= True;
             with drec do
                       begin
                       KeyIndex:= skey1.DynAttr.KeyIndex;
                       DynStyle:= Byte(kinfo.BasicInfo.BasicPart.Style) or $c0;
                       Backgnd:= kinfo.BasicInfo.BasicPart.Backgnd;
                       Foregnd:= kinfo.BasicInfo.BasicPart.Foregnd;
                       Context:= kinfo.BasicInfo.BasicPart.ContextNumber;
                       Cursor:= kinfo.BasicInfo.BasicPart.Cursor;
                       StartKLen:= skey2.Pos-skey1.Pos;
                       StopKLen:= StartKLen;
                       Level:= CurrentLevel+1;
                       KeyIndex[Level]:= keyw;
                       end;

             smemo.SetDynStyleP(skey1, skey2, drec, True, False);
             end
           end
         end;
    skey1.Free;
    skey2.Free;
    nbkeywtodo:= 0;
    end;  { local proc. applykwfound }


begin    { ApplyKeywordsList }
  if Keywords<>nil then kcount:= Keywords.Count
                   else kcount:= 0;

  if Active and (kcount>0) then
    begin
      smemo:= TPPlusMemo(Start.fPMemo);
      kwsome:= False;
      pp:= Start.Par;
      tup:= nil;
      startp:= start.fParNb; stopp:= Stop.ParNumber;

      for i:= startp to stopp do
        begin
          if pp=nil then pp:= smemo.fParagraphs.Pointers[i];

          if i=startp then
            begin
              startoff:= Start.fOffset;
              with fKeywords do
                if startoff>fLongestKeyword+1 then tstartoffset:= startoff-fLongestKeyword-1
                                              else tstartoffset:= 0
            end
          else begin startoff:= 0; tstartoffset:= 0 end;

          if i=stopp then
            begin
              stopoff:= Stop.fOffset;
              with fKeywords do
                  if stopoff+fLongestKeyword<pp^.Length then tlen:= stopoff+fLongestKeyword-tstartoffset
                                                        else tlen:= pp^.Length-tstartoffset
            end
          else begin stopoff:= pp^.Length; tlen:= pp^.Length-tstartoffset end;

          if tlen=0 then begin pp:= nil; Continue end;

          { prepare tlow, tup }
          tlow:= pp^.Text+tstartoffset;
          slowcar:= tlow[tlen];
          supcar:= #0;   { to avoid a warning }
          scar:= #0;

          if Keywords.SomeUpperCase then
            begin
              tup:= smemo.GetUptext(pp, i, tstartoffset, tlen);
              supcar:= tup[tlen]
            end;

          tlow[tlen]:= #0;
          if tup<>nil then tup[tlen]:= #0;

          { tlow, tup are ready, now iterate over keywords }
          nbkeywtodo:= 0;
          for j:= 0 to Keywords.Count-1 do
            begin
              // implement scope-priority check at this stage
              dinfop:= pExtKeyInfoLen(Keywords.Objects[j]);
              slen:= dinfop^.BasicInfo.KeyLen;
              if dinfop^.BasicInfo.BasicPart.ContextNumber<>0 then
                begin
                  if woMatchCase in dinfop^.BasicInfo.BasicPart.Options then tsearch:= tlow
                                                                        else tsearch:= tup;

                  if  (startoff>tstartoffset+slen) then
                              tfound:= tsearch+(startoff-tstartoffset-slen)
                  else tfound:= tsearch;

                  endword:= stopoff+slen-tstartoffset;
                  if endword<tlen then
                    begin
                    scar:= tsearch[endword];
                    tsearch[endword]:= #0
                    end;
                  s:= Keywords.keylist[j];

                  while tfound<>nil do
                    begin
                    tfound:= StrPos(tfound, s);
                    if tfound<>nil then
                       begin
                       if not (woWholeWordsOnly in dinfop^.BasicInfo.BasicPart.Options) or
                          (((tfound=tsearch) or (tlow[tfound-tsearch-1] in fDelimiters)) and
                           (((tfound-tsearch+slen=tlen)       or
                           (tlow[tfound-tsearch+slen] in fDelimiters))))
                           then
                           begin { add to keywtodo array }
                           // implement confirmation here
                           if nbkeywtodo=16 then applykwfound;
                           with keywtodo[nbkeywtodo] do
                             begin
                             Offset:= tfound-tsearch+tstartoffset;
                             keyw:= j;
                             keylen:= slen
                             end;
                           Inc(nbkeywtodo);
                           end;  { boundaries are ok }

                       tfound:= tfound + slen
                       end   { if tfound<>nil }

                    end; { while tfound<>nil }
                  if endword<tlen then tsearch[endword]:= scar;
                end;  { valid word options }
            end;   { j loop over Keyword list }

          if nbkeywtodo>0 then applykwfound;
          tlow[tlen]:= slowcar;
          if tup<>nil then tup[tlen]:= supcar;
          pp:= nil;
        end;     { i loop over paragraphs }

      if kwsome then InvalidateNavs(smemo.fNavigators, Start.Pos, stopp);
    end;   { keyword list not empty }

  if fSubHighlighter<>nil then TCustomExtHighlighter(fSubHighlighter).ApplyKeywordsList(Start, Stop, BaseIndex+kcount)
end;  { method ApplyKeywordsList }


function TCustomExtHighlighter.FindStop(Start, Stop: TPlusNavigator; BaseIndex: Integer): Boolean; { they must be in the same par. }
  var skey1      : TPlusNavigator;
      skeyindex  : SmallInt;
      sssindex   : SmallInt;
      slen       : PlusCardinal;
      t, keyfound: PChar;
      backstart,
      backend,
      eoff       : Longint;
      i          : Integer;
      tlen       : PlusCardinal;
      dinfo      : DynInfoRec;
      leveltofind: SmallInt;

      rightchar, endchar          : Char;
      leftdelcheck, rightdelcheck : Boolean;

      schecklen  : Integer;
      otherindex : Integer;
      otherssinfo,
      pssinfo    : pExStartStopInfo;

      slevel     : SmallInt;
      scount     : Integer;
      spstartdyn : pDynInfoRec;
      smemo      : TPPlusMemo;

begin
  Result:= False;
  spstartdyn:= Start.pDynAttr;
  slevel:= getlevel(spstartdyn^);
  if slevel<0 then Exit;
  skeyindex:= spstartdyn^.KeyIndex[slevel];
  sssindex:= (skeyindex and $7fff) - BaseIndex;
  if sssindex<0 then Exit;

  smemo:= TPPlusMemo(Start.fPMemo);
  if (not Active) or (skeyindex>0) or (fStartStopKeys=nil) or (sssindex>= fStartStopKeys.Count) then
    begin
      if fStartStopKeys=nil then scount:= 0
                            else scount:= fStartStopKeys.Count;
      if fSubHighlighter<>nil then Result:= TCustomExtHighlighter(fSubHighlighter).FindStop(Start, Stop, scount+BaseIndex);
      Exit
    end;

  sKey1:= smemo.fssnav2;
  pSSInfo:= fStartStopKeys.Pointers[sssindex];

  slen:= pSSInfo^.BasicSSInfo.StopLen;
  leftdelcheck:= pSSInfo^.BasicSSInfo.StopLeftCheck;
  rightdelcheck:= pSSInfo^.BasicSSInfo.StopRightCheck;
  schecklen:= slen;
  if leftdelcheck then Inc(schecklen);
  if Start.fOffset>=schecklen then backstart:= schecklen
                              else
                                begin
                                  backstart:= Start.fOffset;
                                  leftdelcheck:= False
                                end;

  { go past the start key }
  if Start.fDynNb>0 then
    with spstartdyn^ do
      if Start.fOffset<=DynOffset+StartKLen then
        begin
          backstart:= DynOffset+StartKLen-Start.fOffset;
          if leftdelcheck and (backstart<Start.fOffset) then Inc(backstart)
        end;

  schecklen:= slen;
  if rightdelcheck then Inc(schecklen);
  if Stop.ParOffset+schecklen>Stop.fPar^.Length then
    begin
      eoff:= Start.fPar^.Length;
      backend:= 0;
      rightdelcheck:= False;
    end
  else
    begin
      eoff:= Stop.fOffset;
      backend:= schecklen
    end;

  tlen:= eoff-Start.ParOffset+backstart+backend;
  if not (woMatchCase in pssinfo^.BasicSSInfo.Attributes.Options) then
      t:= smemo.GetUpText(Start.fPar, Start.fParNb, Start.fOffset-backstart, tlen)
  else
      t:= Start.fPar^.Text+(Start.fOffset-backstart);

  endchar:= #0;

  if t<>nil then
    begin
      endchar:= t[tlen];
      t[tlen]:= #0
    end;
  rightchar:= #0;  { to avoid a warning }

  { t is ready; look for a stop key in it }
  keyfound:= t;
  if leftdelcheck then Inc(keyfound);
  while (not Result) and (keyfound<>nil) do
    begin
      if rightdelcheck then
        begin
          rightchar:= t[tlen-1];
          t[tlen-1]:= #5
        end;

      if pSSInfo^.Scope=EndAtDelScope then
        begin
          while (keyfound^<>#0) and (not (keyfound^ in Delimiters)) do Inc(keyfound);
          Result:= keyfound^<>#0
        end

      else
        begin
          keyfound:= StrPos(keyfound, pSSInfo^.BasicSSInfo.StopKey);
          if rightdelcheck then t[tlen-1]:= rightchar;

          if keyfound<>nil then
             if (pSSInfo^.BasicSSInfo.StopLeftCheck and (keyfound-t>1) and (not (t[keyfound-t-1] in fDelimiters))) or
                (pSSInfo^.BasicSSInfo.StopRightCheck and (keyfound-t<tlen-slen) and (not (t[keyfound-t+slen] in fDelimiters))) then
                   Inc(keyfound)
             else Result:= True
        end;

      if Result or (PSSInfo^.BasicSSInfo.ParStop and (Stop.ParOffset=Stop.fPar^.Length)) then
        begin
          Result:= True;
          if t<>nil then t[tlen]:= endchar;
          sKey1.Assign(Stop);
          if keyfound<>nil then
            begin
              eoff:= Start.Pos-backstart+(keyfound-t)+slen;
              Stop.Pos:= eoff-slen;
              Stop.RightOfDyn;
              while Stop.ForwardToDyn(eoff) do Stop.RemoveDyn;
              Stop.Pos:= eoff
            end;
          if Assigned(fOnStop) then fOnStop(Self, Start, Stop, sssIndex, Result);
          if not Result then
            begin
              if keyfound<>nil then Inc(keyfound);
              Stop.Assign(skey1)
            end
        end;
    end;

  if t<>nil then t[tlen]:= endchar;

  if Result then
    begin
      skey1.Assign(Start);
      skey1.BackToDyn(0);
      if pSSInfo^.BasicSSInfo.StopLeftCheck then Inc(slen);
      skey1.Par^.DynBuffer^[skey1.fDynNb].StopKLen:= slen;

      { check whether another identical start key exists }
      otherindex:= -1;
      with StartStopKeys do
        for i:= (skeyindex and $7fff)+1 to Count-1 do
          with pExStartStopInfo(Pointers[i])^ do
            if (BasicSSInfo.Previous=skeyindex and $7fff) and (Scope= PSSInfo^.Scope) and (Priority=PSSInfo^.Priority) then
             begin
               otherindex:= i;
               Break
             end;

      if otherindex=-1 then
        begin
          leveltofind:= Start.DynAttr.Level-1;
          skey1.Assign(Start);
          while getlevel(skey1.DynAttr)>leveltofind do skey1.BackToDyn(0);
          dinfo:= skey1.DynAttr
        end
      else
        with dinfo,StartStopKeys do
          begin
            otherssinfo:= Pointers[otherindex];
            KeyIndex:= Start.DynAttr.KeyIndex;
            Level:= Start.DynAttr.Level;
            DynStyle:= Byte(otherssinfo^.BasicSSInfo.Attributes.Style) or $c0;
            KeyIndex[Level]:= -32768+otherindex;
            Cursor:= otherssinfo^.BasicSSInfo.Attributes.Cursor;
            Backgnd:= otherssinfo^.BasicSSInfo.Attributes.Backgnd;
            Foregnd:= otherssinfo^.BasicSSInfo.Attributes.Foregnd
          end;

      dinfo.StartKLen:= 0;
      Stop.AddDyn(dinfo);
    end;
end;     { FindStop }


function TCustomExtHighlighter.FindStart(Start, Stop: TPlusNavigator; BaseIndex: Integer): Boolean;
  var currentkey    : SmallInt;
      sKey          : TPlusNavigator;
      slen, tlen    : PlusCardinal;
      t, tup        : PChar;
      tcomp1, tcomp2: PChar;
      backstart,
      backend, eoff : Longint;
      i, ilim       : PlusCardinal;
      j, jlim, jadd : SmallInt;
      jssinfo       : pExStartStopInfo;
      m, mlim       : PlusCardinal;
      dinfo         : DynInfoRec;

      checklen : PlusCardinal;
      smemo    : TPPlusMemo;
      scount   : Integer;
      sdynbuf  : pDynArray;
      currentdyn: DynInfoRec;
      currentpriority, currentscope: SmallInt;

label jContinue, FindHStart;

begin
  Start.RightOfDyn;
  Result:= False;

  if getlevel(Start.pDynAttr^)>=15 then Exit;

  smemo:= TPPlusMemo(Start.fPMemo);
  if fStartStopKeys=nil then scount:= 0
                        else scount:= fStartStopKeys.Count;

  if (not Active) or (scount=0) then goto FindHStart;

  with StartStopKeys do
    begin
      if not fDelChecked then
         for i:= 0 to Count-1 do
           with pExStartStopInfo(Pointers[i])^.BasicSSInfo do
             begin
               StartLeftCheck:= (woWholeWordsOnly in Attributes.Options) and (not (StartKey[0] in fDelimiters));
               StartRightCheck:= (woWholeWordsOnly in Attributes.Options) and (not (StartKey[StartLen-1] in fDelimiters));
               StopLeftCheck:= (StopLen>0) and (woWholeWordsOnly in Attributes.Options) and (not (StopKey[0] in fDelimiters));
               StopRightCheck:= (StopLen>0) and (woWholeWordsOnly in Attributes.Options) and
                                (not (StopKey[StopLen-1] in fDelimiters))
             end;
      fDelChecked:= True
    end;

  sKey:= smemo.fssnav1;
  sKey.Assign(Start);

  if Start.fOffset>=fStartStopKeys.fLongestStartKey+1 then backstart:= fStartStopKeys.fLongestStartKey+1
                                                      else backstart:= Start.ParOffset;
  skey.Pos:= Start.Pos-backstart;

  eoff:= Stop.ParOffset;
  if eoff+fStartStopKeys.fLongestStartKey+1>Stop.fPar^.Length then backend:= Stop.fPar^.Length-eoff
                                                              else backend:= fStartStopKeys.fLongestStartKey+1;
  tlen:= eoff-Start.fOffset+backstart+backend;
  if tlen<=0 then goto FindHStart;

  t:= Start.fPar^.Text + skey.fOffset;
  tup:= nil;

  m:= skey.DynNb;
  mlim:= Start.fPar^.DynCount;

  i:= 0;
  jadd:= fStartStopKeys.fElementSpace;
  jlim:= fStartStopKeys.Count;
  sdynbuf:= Start.fPar^.DynBuffer;
  ilim:= tlen-backend;
  slen:= 0;

  currentdyn:= skey.DynAttr;
  currentpriority:= getpriority(currentdyn, BaseIndex);
  currentscope:= dyntoscope(currentdyn);

  while i<=ilim do
    begin
      while (m<mlim) and (sdynbuf^[m].DynOffset<=i+skey.fOffset) do
        begin
          currentdyn:= sdynbuf^[m];
          Inc(m);
          currentpriority:= GetPriority(currentdyn, BaseIndex);
          currentscope:= DynToScope(currentdyn)
        end;

      jssinfo:= fStartStopKeys.Pointers[0];
      j:= 0;
      while j<jlim do
        begin
          if (jssinfo^.Priority=QueryPriority) or
             ((jssinfo^.Scope=0) and (jssinfo^.Priority>currentpriority)) or
             ((jssinfo^.Scope=currentscope) and ((jssinfo^.Scope<>0) or (currentdyn.DynStyle and $80=0))) then
            begin
              slen:= jssinfo^.BasicSSInfo.StartLen;
              if jssinfo^.BasicSSInfo.StartRightCheck then checklen:= slen+1 else checklen:= slen;
              if (i+checklen>backstart) and (i+slen<=tlen) then
                begin
                  tcomp1:= jssinfo^.BasicSSInfo.StartKey;
                  if not (woMatchCase in jssinfo^.BasicSSInfo.Attributes.Options) then
                    begin
                      if tup=nil then tup:= smemo.GetUpText(Start.fPar, Start.fParNb, skey.fOffset, tlen);
                      tcomp2:= tup+i
                    end
                  else tcomp2:= t+i;

                  while (slen>0) and (tcomp1^=tcomp2^) do
                    begin
                      Dec(slen);
                      Inc(tcomp1);
                      Inc(tcomp2)
                    end;

                  if slen=0 then
                    begin
                      slen:= jssinfo^.BasicSSInfo.StartLen;
                      if not (jssinfo^.BasicSSInfo.StartLeftCheck and (i>0) and (not (t[i-1] in fDelimiters))) and
                         not(jssinfo^.BasicSSInfo.StartRightCheck and (i+slen<tlen) and (not (t[i+slen] in fDelimiters))) then
                        begin
                          Result:= True;
                          Break
                        end
                    end
                end
            end;    // scope, priority is ok
          jContinue:
          Inc(PChar(jSSInfo), jadd);
          Inc(j)
        end;  // j loop over start/stop keys

      if not Result then Inc(i)
      else
        begin
          SKey.Pos:= SKey.Pos+i;
          sKey.RightOfDyn;
          if Assigned(fOnStart) then fOnStart(Self, Skey, nil, j, Result);
          if not Result then goto jContinue;

          Stop.Assign(skey);
          dinfo.Level:= getlevel(Stop.DynAttr)+1;
          { check whether we are nesting within the same kind of start-stop key,
            which must be avoided if stopkey=startkey}
          if dinfo.Level-1>=0 then
            if (skey.DynAttr.KeyIndex[dinfo.Level-1]=j or SmallInt($8000)) and
                ((jssinfo^.BasicSSInfo.StartLen=jssinfo^.BasicSSInfo.StopLen) and
                 (StrComp(jssinfo^.BasicSSInfo.StartKey, jssinfo^.BasicSSInfo.StopKey)=0)) then
              begin
                Inc(i);
                Result:= False;
                Continue
              end;
          if dinfo.Level>0 then
            with Stop do
              begin
                if BackToDyn(0) then Par^.DynBuffer^[DynNb].StopKLen:= 0;
                Assign(skey); RightOfDyn
              end;

          {remove stray dyn codes}
          while Stop.ForwardToDyn(sKey.Pos+slen) do Stop.RemoveDyn;

          with dinfo do
            begin
              DynOffset:= skey.ParOffset;
              KeyIndex:= Stop.DynAttr.KeyIndex;
              DynStyle:= Byte(jssinfo^.BasicSSInfo.Attributes.Style) or $c0;
              KeyIndex[Level]:= -32768+j + BaseIndex;
              Cursor:= jssinfo^.BasicSSInfo.Attributes.Cursor;
              Backgnd:= jssinfo^.BasicSSInfo.Attributes.Backgnd;
              Foregnd:= jssinfo^.BasicSSInfo.Attributes.Foregnd;
              Context:= jssinfo^.BasicSSInfo.Attributes.ContextNumber;
              StartKlen:= slen;
              StopKLen:= jssinfo^.BasicSSInfo.StopLen;
              if jssinfo^.BasicSSInfo.StartRightCheck then Inc(StartKLen)
            end;
          with skey do
            if (fDynNb<fPar^.DynCount) and (fPar^.DynBuffer^[fDynNb].DynOffset=fOffset) then
                     fPar^.DynBuffer^[fDynNb]:= dinfo
            else AddDyn(dinfo);
          Stop.Pos:= sKey.Pos+slen;
          Stop.RightOfDyn;
          Break
        end { found start key }

      end; { iloop in text buffer }

  FindHStart:
  if fSubHighlighter<>nil then Result:= Result or TCustomExtHighlighter(fSubHighlighter).FindStart(Start, Stop, scount+BaseIndex)
end;    { FindStart }

end.
