unit ExtKeys;

interface

uses WinTypes, WinProcs, Classes, Graphics, Forms, Controls, Buttons,
  StdCtrls, Grids, ExtCtrls, ExtHilit, Dialogs, DsgnIntf;

type TExtKeywordsProperty = class(TStringProperty)
                             public
                             function GetAttributes: TPropertyAttributes; override;
                             function GetValue: string; override;
                             procedure Edit; override;
                             end;

     TExtStartStopProperty = class(TStringProperty)
                             public
                             function GetAttributes: TPropertyAttributes; override;
                             function GetValue: string; override;
                             procedure Edit; override;
                             end;


  TDlgExtKeysEditor = class(TForm)
    OKBtn: TBitBtn;
    CancelBtn: TBitBtn;
    gbTitle: TGroupBox;
    StrGKeys: TStringGrid;
    gbOptions: TGroupBox;
    ChkBMatchCase: TCheckBox;
    ChkBWholeWords: TCheckBox;
    ChkBHighlight: TCheckBox;
    ChkBAltFont: TCheckBox;
    ChkBItalic: TCheckBox;
    ChkBBold: TCheckBox;
    ChkBUnderline: TCheckBox;
    ChkBStrikeOut: TCheckBox;
    LblHint: TLabel;
    ColorDialog1: TColorDialog;
    ShpForegnd: TShape;
    BtnForegnd: TButton;
    BtnBackgnd: TButton;
    ShpBackgnd: TShape;
    BtnResetColors: TButton;
    Label1: TLabel;
    EdContext: TEdit;
    BtnRemove: TButton;
    LblMouse: TLabel;
    cbCursors: TComboBox;
    ChkBCrossPar: TCheckBox;
    Label2: TLabel;
    EdScope: TEdit;
    Label3: TLabel;
    EdPriority: TEdit;
    procedure FormShow(Sender: TObject);
    procedure StrGKeysSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure StrGKeysSetEditText(Sender: TObject; ACol, ARow: Longint;
      const Value: String);
    procedure ChkBxClick(Sender: TObject);
    procedure BtnRemoveClick(Sender: TObject);
    procedure StrGKeysEnter(Sender: TObject);
    procedure StrGKeysExit(Sender: TObject);
    procedure StrGKeywordsGetEditText(Sender: TObject; ACol, ARow: Longint;
      var Value: OpenString);
    procedure StrGKeysKeyPress(Sender: TObject; var Key: Char);
    procedure BtnForegndClick(Sender: TObject);
    procedure BtnBackgndClick(Sender: TObject);
    procedure BtnResetColorsClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure cbCursorsChange(Sender: TObject);
    procedure EdContextExit(Sender: TObject);
  private
    { Private declarations }
    inSelect: Boolean;
    Modified: Boolean;
    woList: TList;
    procedure AddCursor(const s: string);
  public
    { Public declarations }
    {class function EditKeywordList(KeywordList: TKeywordList): Boolean;}
  end;

var
  DlgExtKeysEditor: TDlgExtKeysEditor;

function EditExtKeywordList(KeywordList: TExtKeywordList): Boolean;
function EditExtStartStopList(StartStopList: TExtStartStopList): Boolean;

procedure Register;

implementation
uses SysUtils, Messages, PlusMemo;
{$R *.DFM}

//type pExtWordInfo = ^TExtKeywordInfo;

function EditExtKeywordList(KeywordList: TExtKeywordList): Boolean;
  var i: Integer;
      pk: pExtKeyInfoLen;
      kw: string;
  begin
  with TDlgExtKeysEditor.Create(Application) do
    begin
    woList:= TList.Create;
    StrGKeys.RowCount:= KeywordList.Count+1;

    for i:= 0 to KeywordList.Count-1 do
      begin
      new(pk);
      pk^:= pExtKeyInfoLen(KeywordList.Objects[i])^;
      woList.Add(pk);
      StrGKeys.Cells[0, i]:= KeywordList[i]
      end;
    new(pk);
    if KeywordList.Count>0 then pk^:= pExtKeyInfoLen(KeywordList.Objects[KeywordList.Count-1])^
    else with pk^, pk^.BasicInfo.BasicPart do
           begin
           Options:= [];
           Style:= [];
           ContextNumber:= 1;
           Backgnd:= 0; Foregnd:= 0;
           Scope:= 0; Priority:= 0;
           Cursor:= crDefault
           end;
    woList.Add(pk);

    if (ShowModal=mrOk) and Modified then
      with Keywordlist do
        begin
        Clear;
        BeginUpdate;
        for i:= 0 to StrGKeys.RowCount-2 do
          begin
          kw:= StrGKeys.Cells[0,i];
          if kw<>'' then AddObject(kw, woList[i])
                    else Dispose(pExtKeyInfoLen(woList[i]))
          end;
        EndUpdate;
        Result:= True
      end
    else begin
         for i:= 0 to woList.Count-1 do Dispose(pExtKeyInfoLen(woList[i]));
         Result:= False
         end;
    woList.Free;
    Free
    end
  end;

function EditExtStartStopList(StartStopList: TExtStartStopList): Boolean;
  var i: Integer;
      pk: pExtKeyInfoLen;
      ssinfo: pExStartStopInfo;
      ps: Boolean;
      startkey: string;
  begin
  with TDlgExtKeysEditor.Create(Application) do
    begin
    Caption:= 'Extended Start-Stop keys Editor';
    StrGKeys.ColCount:= 2;
    StrGKeys.DefaultColWidth:= 130;
    gbTitle.Caption:= '   Start key                    Stop key';
    gbOptions.Caption:= 'Options for selected start-stop key';
    ChkBCrossPar.Show;

    woList:= TList.Create;
    StrGKeys.RowCount:= StartStopList.Count+1;

    for i:= 0 to StartStopList.Count-1 do
      begin
      new(pk);
      ssinfo:= StartStopList.Pointers[i];
      pk^.BasicInfo.BasicPart:= ssinfo^.BasicSSInfo.Attributes;
      pk^.Scope:= ssinfo^.Scope;
      pk^.Priority:= ssinfo^.Priority;
      woList.Add(pk);
      if ssinfo^.BasicSSInfo.ParStop then Byte(pk^.BasicInfo.BasicPart.Options):= Byte(pk^.BasicInfo.BasicPart.Options) or $80;
      StrGKeys.Cells[0, i]:= StrPas(ssinfo^.BasicSSInfo.StartKey);
      StrGKeys.Cells[1, i]:= StrPas(ssinfo^.BasicSSInfo.StopKey)
      end;
    new(pk);
    if StartStopList.Count>0 then
      begin
      ssinfo:= StartStopList.Pointers[StartStopList.Count-1];
      pk^.BasicInfo.BasicPart:= ssinfo^.BasicSSInfo.Attributes;
      pk^.Scope:= ssinfo^.Scope;
      pk^.Priority:= ssinfo^.Priority;
      if ssinfo^.BasicSSInfo.ParStop then Byte(pk^.BasicInfo.BasicPart.Options):= Byte(pk^.BasicInfo.BasicPart.Options) or $80;
      end
    else with pk^, pk^.BasicInfo.BasicPart do
           begin
           Options:= [];
           Style:= [];
           ContextNumber:= 1;
           Backgnd:= -1; Foregnd:= -1;
           Scope:= 0; Priority:= 0;
           Cursor:= crDefault
           end;
    woList.Add(pk);

    if (ShowModal=mrOk) and Modified then
      with StartStopList do
        begin
        Clear;
        for i:= 0 to StrGKeys.RowCount-2 do
          begin
          pk:= woList[i];
          ps:= Byte(pk^.BasicInfo.BasicPart.Options) and $80<>0;
          Byte(pk^.BasicInfo.BasicPart.Options):= Byte(pk^.BasicInfo.BasicPart.Options) and $7f;
          startkey:= StrGKeys.Cells[0,i];
          if startkey<>'' then
            AddExStartStopKey(startkey, StrGKeys.Cells[1, i],
                            pk^.BasicInfo.BasicPart.Options, pk^.BasicInfo.BasicPart.Style, pk^.BasicInfo.BasicPart.ContextNumber,
                            pk^.Scope, pk^.Priority,
                            pk^.BasicInfo.BasicPart.Cursor, pk^.BasicInfo.BasicPart.Backgnd, pk^.BasicInfo.BasicPart.Foregnd, ps);
          Dispose(pk)
          end;
        Result:= True
        end
    else begin
         for i:= 0 to woList.Count-1 do Dispose(pExtKeyInfoLen(woList[i]));
         Result:= False
         end;
    woList.Free;
    Free
    end
  end;

function TExtKeywordsProperty.GetAttributes;
  begin
  Result:= [paDialog, paAutoUpdate]
  end;

function TExtKeywordsProperty.GetValue: string;
  begin
  Result:= '(TKeywordList)'
  end;

procedure TExtKeywordsProperty.Edit;
  var MyMemo: TExtHighlighter;
  begin
    MyMemo:= GetComponent(0) as TExtHighlighter;
    if EditExtKeywordList(MyMemo.Keywords) then
      begin
      MyMemo.ReApplyKeys;
      Modified
      end
  end;

function TExtStartStopProperty.GetAttributes;
  begin
  Result:= [paDialog, paAutoUpdate]
  end;

function TExtStartStopProperty.GetValue: string;
  begin
  Result:= '(TExtStartStopKeyList)'
  end;

procedure TExtStartStopProperty.Edit;
  var MyMemo: TExtHighlighter;
  begin
    MyMemo:= GetComponent(0) as TExtHighlighter;
    if EditExtStartStopList(MyMemo.StartStopKeys) then
      begin
      MyMemo.ReApplyKeys;
      Modified
      end
  end;



procedure TDlgExtKeysEditor.FormShow(Sender: TObject);
var dummy: Boolean;
begin
  StrGKeysSelectCell(Self, 0, StrGKeys.Row, dummy)
end;

procedure TDlgExtKeysEditor.StrGKeysSelectCell(Sender: TObject; Col,
  Row: Longint; var CanSelect: Boolean);
begin
CanSelect:= True;
inSelect:= True;   { avoid OnClick updating info }
with pExtKeyInfoLen(woList[Row])^ do
  begin
  ChkBMatchCase.Checked:= woMatchCase in BasicInfo.BasicPart.Options;
  ChkBWholeWords.Checked:= woWholeWordsOnly in BasicInfo.BasicPart.Options;
  ChkBBold.Checked:= fsBold in BasicInfo.BasicPart.Style;
  ChkBItalic.Checked:= fsItalic in BasicInfo.BasicPart.Style;
  ChkBUnderline.Checked:= fsUnderline in BasicInfo.BasicPart.Style;
  ChkBStrikeOut.Checked:= fsStrikeOut in BasicInfo.BasicPart.Style;
  ChkBHighlight.Checked:=  Byte(BasicInfo.BasicPart.Style) and (1 shl Ord(fsHighlight)) <> 0;
  ChkBAltFont.Checked:= Byte(BasicInfo.BasicPart.Style) and (1 shl Ord(fsAltFont)) <> 0;
  ChkBCrossPar.Checked:= (Byte(BasicInfo.BasicPart.Options) and $80) = 0;
  if (BasicInfo.BasicPart.Backgnd=0) and (BasicInfo.BasicPart.Foregnd=0) then begin ShpForegnd.Hide; ShpBackgnd.Hide end
  else begin
       ShpForegnd.Brush.Color:= BasicInfo.BasicPart.Foregnd;
       ShpBackgnd.Brush.Color:= BasicInfo.BasicPart.Backgnd;
       ShpForegnd.Visible:= BasicInfo.BasicPart.Foregnd<>-1;
       ShpBackgnd.Visible:= BasicInfo.BasicPart.Backgnd<>-1
       end;
  EdContext.Text:= IntToStr(BasicInfo.BasicPart.ContextNumber);
  EdScope.Text  := IntToStr(Scope);
  EdPriority.Text:= IntToStr(Priority);
  EdContext.Modified:= False;
  EdScope.Modified:= False;
  EdPriority.Modified:= False;
  LblMouse.Cursor:=BasicInfo.BasicPart.Cursor;
  cbCursors.ItemIndex:= cbCursors.Items.IndexOf(CursorToString(BasicInfo.BasicPart.Cursor))
  end;

inSelect:= False
end;

procedure TDlgExtKeysEditor.StrGKeysSetEditText(Sender: TObject; ACol,
  ARow: Longint; const Value: String);
begin
Modified:= True;
LblHint.Show;
end;

procedure TDlgExtKeysEditor.ChkBxClick(Sender: TObject);
begin
if not inSelect then
  with pExtKeyInfoLen(woList[StrGKeys.Row])^.BasicInfo.BasicPart do
    begin
    Modified:= True;
    if ChkBMatchCase.Checked then Options:= [woMatchCase]
                             else Options:= [];
    if ChkBWholeWords.Checked then Options:= Options + [woWholeWordsOnly];
    if ChkBBold.Checked then Style:= [fsBold]
                        else Style:= [];
    if ChkBItalic.Checked then Style:= Style + [fsItalic];
    if ChkBUnderline.Checked then Style:= Style + [fsUnderline];
    if ChkBStrikeOut.Checked then Style:= Style + [fsStrikeOut];
    if ChkBHighlight.Checked then Byte(Style):= Byte(Style) or (1 shl Ord(fsHighlight));
    if ChkBAltFont.Checked then Byte(Style):= Byte(Style) or (1 shl Ord(fsAltFont));
    if not ChkBCrossPar.Checked then Byte(Options):= Byte(Options) or $80
                                else Byte(Options):= Byte(Options) and $7f
    end;
end;

procedure TDlgExtKeysEditor.BtnRemoveClick(Sender: TObject);
var i: Integer;  dummy: Boolean;
var pk: pExtKeyInfoLen;
begin
with StrGKeys do
  if Row<RowCount-1 then
    begin
    pk:= woList[Row];
    Dispose(pk);
    for i:= Row to RowCount-2 do
        begin
        Rows[i]:= Rows[i+1];
        woList[i]:= woList[i+1];
        end;
    RowCount:= RowCount-1;
    woList.Count:= woList.Count-1;
    StrGKeysSelectCell(Self, 0, Row, dummy);
    Modified:= True
    end
end;

procedure TDlgExtKeysEditor.StrGKeysEnter(Sender: TObject);
begin
LblHint.Show
end;

procedure TDlgExtKeysEditor.StrGKeysExit(Sender: TObject);
begin
LblHint.Hide
end;

procedure TDlgExtKeysEditor.StrGKeywordsGetEditText(Sender: TObject; ACol,
  ARow: Longint; var Value: OpenString);
begin
{LblHint.Hide }
end;

procedure TDlgExtKeysEditor.StrGKeysKeyPress(Sender: TObject; var Key: Char);
var pk: pExtKeyInfoLen;
begin
with StrGKeys do
  if Row=RowCount-1 then
    begin
    RowCount:= RowCount+1;
    new(pk);
    woList.Add(pk);
    pk^:= pExtKeyInfoLen(woList[Row])^
    end;

end;


procedure TDlgExtKeysEditor.BtnForegndClick(Sender: TObject);
var saved: TColor;
begin
if ShpForegnd.Visible then ColorDialog1.Color:= ShpForegnd.Brush.Color
                      else ColorDialog1.Color:= clBlack;
saved:= ColorDialog1.Color;
if ColorDialog1.Execute then
  if ColorDialog1.Color<>saved then
    begin
    ShpForegnd.Brush.Color:= ColorDialog1.Color;
    ShpForegnd.Show;
    with pExtKeyInfoLen(woList[StrGKeys.Row])^.BasicInfo.BasicPart do
      begin
      Foregnd:= ColorDialog1.Color;
      //ShpBackgnd.Brush.Color:= Backgnd;
      //ShpBackgnd.Show
      end;
    Modified:= True
    end
end;

procedure TDlgExtKeysEditor.BtnBackgndClick(Sender: TObject);
var saved: TColor;
begin
if ShpBackgnd.Visible then ColorDialog1.Color:= ShpBackgnd.Brush.Color
                      else ColorDialog1.Color:= clWhite;
saved:= ColorDialog1.Color;
if ColorDialog1.Execute then
  if ColorDialog1.Color<>saved then
    begin
    ShpBackgnd.Brush.Color:= ColorDialog1.Color;
    ShpBackgnd.Show;
    with pExtKeyInfoLen(woList[StrGKeys.Row])^.BasicInfo.BasicPart do
      begin
      Backgnd:= ColorDialog1.Color;
      //ShpForegnd.Brush.Color:= Foregnd;
      //ShpForegnd.Show
      end;
    Modified:= True
    end
end;



procedure TDlgExtKeysEditor.BtnResetColorsClick(Sender: TObject);
begin
if ShpBackgnd.Visible then
  begin
  with pExtKeyInfoLen(woList[StrGKeys.Row])^.BasicInfo.BasicPart do begin Backgnd:= -1; Foregnd:= -1 end;
  Modified:= True;
  ShpBackgnd.Hide;
  ShpForegnd.Hide;
  BtnResetColors.Enabled:= False
  end
end;

procedure TDlgExtKeysEditor.FormCreate(Sender: TObject);
begin
GetCursorValues(AddCursor);
end;

procedure TDlgExtKeysEditor.AddCursor(const s: string);
begin
cbCursors.Items.Add(s)
end;

procedure TDlgExtKeysEditor.cbCursorsChange(Sender: TObject);
begin
if not inSelect then
  with pExtKeyInfoLen(woList[StrGKeys.Row])^.BasicInfo.BasicPart do
    begin
    Modified:= True;
    Cursor:= StringToCursor(cbCursors.Items[cbCursors.ItemIndex]);
    LblMouse.Cursor:= Cursor
    end
end;


procedure TDlgExtKeysEditor.EdContextExit(Sender: TObject);
var edit: TEdit; newvalue: SmallInt;
begin
edit:= Sender as TEdit;
if edit.Modified then
  begin
  with pExtKeyInfoLen(woList[StrGKeys.Row])^ do
    begin
    Modified:= True;
    try
    newvalue:= StrToInt(edit.Text);
    if Sender= EdContext then BasicInfo.BasicPart.ContextNumber:= newvalue
    else
      if Sender=EdScope then Scope:= newvalue
                        else Priority:= newvalue
    except
      on E: Exception do
        begin
        ShowMessage(E.Message);
        edit.SetFocus
        end
      end
    end
  end
end;
procedure Register;
begin
  RegisterComponents('ECM', [TExtHighlighter]);
  RegisterPropertyEditor(TypeInfo(TExtKeywordList), TExtHighlighter, 'Keywords', TExtKeywordsProperty);
  RegisterPropertyEditor(TypeInfo(TExtStartStopList), TExtHighlighter, 'StartStopKeys', TExtStartStopProperty);
end;

end.
