unit URLHighlight;
{  Electro-Concept Mauricie, 1998-1999-2000

  Version 5.1b June 6, 2000 release
    Changes since v5.1a, Jan. 2000:
      Added property Priority: Boolean;
        if True, highlighting a url occurs even if it spans dynamic highlighted boundaries
        sections, or if found completely inside of a dynamically highlighted part.
        This enables proper working of TUrlHighlihter with the TPMLiveSpell component.
        In general, if you use TUrlHighlighter as a sub highlighter of TPMLiveSpell, you
        should set Priority to True, otherwise you should leave it to False.

  A component to dynamically highlight Internet urls of all kind, including short forms of
  web and email urls (ex: www.scientifique.com)

  Place a TUrlHighlight on your form, set your TPlusMemo.Highlighter property to point to it,
  et voil, urls are turned colored, underlined, bolded, as you want.

  Property LaunchURLs decides whether clicking on a highlighted url will make the system
  launch it (ex: clicking on  www.scientifique.com  will launch the default browser)

  You can control the highlighting attributes for different kind of urls from within the Object
  Inspector.  This is the purpose of properties WebURL, EmailURL and so on.

  You can also control in which part of your text the urls will be highlighted.  For this, you
  have one design time property and one event:
    - Scope: set it to the value of ContextNumber of the start-stop key where you want urls
             highlighted.  Its default value is zero, meaning that they will be highlighted
             only in the unhighlighted text, i.e. outside of any start-stop key.

    - OnURL(Sender: TObject; CurrentContext: SmallInt; URLType: TURLType): Boolean; this event
            let your application decide whether to highlight the url just found.
              CurrentContext is the value assigned to the start-stop key where the url is found
              (zero if outside of any start-stop key), URLType is the type of url about to be
              highlighted (an enumerated type defined in the interface part below).
              The url is highlighted if and only if your handler returs True.  If your application
              does not provide a handler, only the Scope property is used to decide whether the
              url is highlighted.

   Property BaseContext is used to assign a given ContextNumber to highlighted urls.  This number
   gets passed to TPlusMemo.OnContext event, that your application can use to know when a url has
   been clicked.  Its default value is 1024.  The actual context number given to a url is
   BaseContext + some index, where some index is something in the range 0..14 that depends on
   the kind of url.  See the constants below for their enumeration. }


{ License terms:
----------------
  This component is freeware. You may copy those files AS LONG AS YOU COPY ALL OF THEM and don't charge
  any kind of fee for it. If you want to change the source code in order to improve the components features,
  performance, etc. please send us the new source code so that we can have a look at it. The changed source
  code should contain descriptions what you have changed, and of course your name. The only thing you MAY NOT
  CHANGE is the ORIGINAL COPYRIGHT INFORMATION.

  Disclaimer
----------
   The author is NOT liable for any DAMAGES resulting from the use and misuse of the component, especially
   he is NOT liable for DAMAGES that were caused BY ANY VERSION WHICH HAS NOT BEEN PROGRAMMED BY THE ORIGINAL
   AUTHOR HIMSELF.

Contact us
----------
  If you have questions, suggestions or comments, feel free to send an E-Mail to
  Electro-Concept Mauricie  http://www.ecmqc.com/     mailto:ecm@ecmqc.com

Thank you
  }

interface

uses Classes, Graphics, Controls, Plusmemo;

const
  URL_HTTP     = 0;    { values to add to ContextNumber to know which particular kind of url }
  URL_WWW      = 1;
  URL_MAILTO   = 2;
  URL_MAILshort= 3;
  URL_NEWS     = 4;
  URL_NNTP     = 5;
  URL_FTP      = 6;
  URL_FTPshort = 7;
  URL_GOPHER   = 8;
  URL_GOPHERshort = 9;
  URL_TELNET   = 10;
  URL_TN3270   = 11;
  URL_WAIS     = 12;
  URL_PROSPERO = 13;
  URL_HTTPS    = 14;

type
  TURLType = (urlWeb, urlEmail, urlNNTP, urlFile, urlOther);
  TURLEvent = function (Sender: TObject; CurrentContext: SmallInt; URLType: TURLType): Boolean of object;

  THighlightInfo = class(TPersistent) { Presents highlighting options in Object Inspector }
    private
      fAltFont    : Boolean;
      fStyle      : TFontStyles;
      fBackground,
      fForeground : TColor;
      fCursor     : TCursor;
      fOnChange   : TNotifyEvent;
      procedure setAltFont   (af: Boolean);
      procedure setStyle     (s: TFontStyles);
      procedure setBackground(bg: TColor);
      procedure setForeground(fg: TColor);

    public
      procedure Assign(Source: TPersistent); override;
      property OnChange: TNotifyEvent read fOnChange write fOnChange;

    published
      property AltFont   : Boolean     read fAltFont    write setAltFont;
      property Style     : TFontStyles read fStyle      write setStyle;
      property Background: TColor      read fBackground write setBackground;
      property Foreground: TColor      read fForeground write setForeground;
      property Cursor    : TCursor     read fCursor     write fCursor;
    end;

  TURLHighlighter = class(TPlusHighlighter)
    private
      fWebURL,
      fEmailURL,
      fFileURL,
      fNNTP_URL,
      fOtherURL     : THighlightInfo;
      fBaseContext  : Integer;
      fLaunchURLS   : Boolean;
      fScope        : Integer;
      fPriority     : Boolean;
      fOnUrl        : TURLEvent;

      procedure HighlightChange(Sender: TObject);
      procedure setWebURL(h: THighlightInfo);
      procedure setEmailURL(h: THighlightInfo);
      procedure setFileURL(h: THighlightInfo);
      procedure setNNTP_URL(h: THighlightInfo);
      procedure setOtherURL(h: THighlightInfo);
      procedure setScope(sc: Integer);
      procedure setPriority(const Value: Boolean);

    protected
      procedure Loaded; override;
      procedure Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure ApplyKeywordsList(Start, Stop: TPlusNavigator; BaseIndex: Integer);  override;

    public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;

    published
      property BaseContext: Integer      read fBaseContext write fBaseContext;
      property WebURL   : THighlightInfo read fWebURL   write setWebURL;
      property EmailURL : THighlightInfo read fEmailURL write setEmailURL;
      property NNTP_URL : THighlightInfo read fNNTP_URL write setNNTP_URL;
      property FileURL  : THighlightInfo read fFileURL  write setFileURL;
      property OtherURL : THighlightInfo read fOtherURL write setOtherURL;
      property LaunchURLs: Boolean       read fLaunchURLs write fLaunchURLs;
      property Scope    : Integer        read fScope    write setScope;
      property Priority : Boolean        read fPriority write setPriority;
      property OnURL    : TURLEvent      read fOnUrl    write fOnUrl;
    end;


procedure Register;

implementation

uses SysUtils, Forms, ShellAPI, WinTypes;

const
  WebURLColor    = clBlue;    { default creation colors for TURLHighlighter }
  EmailURLColor  = clGreen;
  NNTP_URLColor  = clRed;
  FileURLColor   = clFuchsia;
  OtherURLColor  = clSilver;

  NbUrls = 15;

type
  URLRecord = record
      Key      : string;
      URLSpecie: TURLType;
      StartURL : Boolean
    end;

  URLArray = array[0..NbURLs-1] of URLRecord;

const
  URLs: URLArray =
      ((Key: 'HTTP://';   URLSpecie: urlWeb; StartURL: True),
       (Key: 'WWW.';      URLSpecie: urlWeb; StartURL: True),
       (Key: 'MAILTO:';   URLSpecie: urlEmail; StartURL: True),
       (Key: '@';         URLSpecie: urlEmail; StartURL: False),
       (Key: 'NEWS:';     URLSpecie: urlNNTP;  StartURL: True),
       (Key: 'NNTP://';   URLSpecie: urlNNTP;  StartURL: True),
       (Key: 'FTP://';    URLSpecie: urlFile;  StartURL: True),
       (Key: 'FTP.';      URLSpecie: urlFile;  StartURL: True),
       (Key: 'GOPHER://'; URLSpecie: urlOther; StartURL: True),
       (Key: 'GOPHER.';   URLSpecie: urlOther; StartURL: True),
       (Key: 'TELNET://'; URLSpecie: urlOther; StartURL: True),
       (Key: 'TN3270://'; URLSpecie: urlOther; StartURL: True),
       (Key: 'WAIS://';   URLSpecie: urlOther; StartURL: True),
       (Key: 'PROSPERO://'; URLSpecie: urlOther; StartURL: True),
       (Key: 'HTTPS://';  URLSpecie: urlWeb;   StartURL: True));

  URLDelimiters = ['<', '>', {':',} ' ', #9, '"', '''', '{', '}', ';'];

var LongestKey: Integer;  { initialized at startup }

procedure THighlightInfo.Assign(Source: TPersistent);
var srch: THighlightInfo;
begin
  if Source is THighlightInfo then
    begin
      srch:= THighlightInfo(Source);
      fAltFont   := srch.fAltFont;
      fStyle     := srch.fStyle;
      fBackground:= srch.fBackground;
      fForeground:= srch.fForeground;
      if Assigned(fOnChange) then fOnChange(Self)
    end
  else inherited Assign(Source)
end;

procedure THighlightInfo.setAltFont(af: Boolean);
begin
  if af<>fAltFont then
    begin
      fAltFont:= af;
      if Assigned(fOnChange) then fOnChange(Self)
    end
end;

procedure THighlightInfo.setStyle(s: TFontStyles);
begin
if s<>fStyle then
  begin
    fStyle:= s;
    if Assigned(fOnChange) then fOnChange(Self)
  end
end;

procedure THighlightInfo.setBackground(bg: TColor);
begin
  if bg<>fBackground then
    begin
      fBackground:= bg;
      if Assigned(fOnChange) then fOnChange(Self)
    end
end;

procedure THighlightInfo.setForeground(fg: TColor);
begin
  if fg<>fForeground then
    begin
      fForeground:= fg;
      if Assigned(fOnChange) then fOnChange(Self)
    end
end;

type TProtPlusMemo = class(TPlusMemo);  { to have access to protected members of TPlusMemo }

constructor TURLHighlighter.Create(AOwner: TComponent);
  procedure CreateHighlight(var h: THighlightInfo; style: TExtFontStyles; bk, fr: TColor);
    begin
      h:= THighlightInfo.Create;
      h.fAltFont:=  TPlusFontStyle(fsAltFont) in TPlusFontStyles(style);
      TPlusFontStyles(h.fStyle):= TPlusFontStyles(style) - [TPlusFontStyle(fsAltFont)];
      h.fBackground:= bk;
      h.fForeground:= fr;
      h.Cursor:= {$IFDEF WIN32} {$IFNDEF VER90} crHandPoint; {$ELSE} crUpArrow;  {$ENDIF} {$ELSE} crUpArrow {$ENDIF}
      h.fOnChange:= HighlightChange
    end;
begin
  inherited Create(AOwner);
  fBaseContext:= 1024;
  CreateHighlight(fWebURL, [fsUnderline], -1, WebURLColor);
  CreateHighlight(fEmailURL, [fsUnderline], -1, EmailURLColor);
  CreateHighlight(fNNTP_URL, [fsUnderline], -1, NNTP_URLColor);
  CreateHighlight(fFileURL, [fsUnderline], -1, FileURLColor);
  CreateHighlight(fOtherURL, [fsItalic], clInfoBk, OtherURLColor);
  fLaunchURLs:= True
end;


destructor TURLHighlighter.Destroy;
begin
  fWebURL.Destroy;
  fEmailURL.Destroy;
  fNNTP_URL.Destroy;
  fFileURL.Destroy;
  fOtherURL.Destroy;
  inherited Destroy
end;

procedure TURLHighlighter.HighlightChange(Sender: TObject);
  { event handler for internal THighlightInfo modifications }
var i: Integer;
begin
  if csLoading in ComponentState then Exit;
  if csDesigning in ComponentState then
    for i:= 0 to MemoList.Count-1 do
        TPlusMemo(MemoList[i]).ReApplyKeywords
end;

procedure TURLHighlighter.Loaded;
  begin
    inherited Loaded;
    HighlightChange(nil)
  end;

procedure TURLHighlighter.Notification(AComponent: TComponent; Operation: TOperation);
var ContextString: string;
begin
  if (Byte(Operation)=pm_Msg) then
    with TPlusMemo(AComponent).NotifyMsg do
      begin
        if (Msg = pm_Context) and LaunchUrls and (WParam>=BaseContext) and (WParam<BaseContext+NbUrls) then
          begin
            ContextString:= pString(LParam)^;
            { the following is necessary to have Win95 properly launch some short form urls
              Win98 and WinNT do not require this }
            case WParam - BaseContext of
              URL_MAILshort  : ContextString:= 'mailto:' + ContextString;
              URL_FTPshort   : ContextString:= 'ftp://' + ContextString;
              URL_GOPHERshort: ContextString:= 'gopher://' + ContextString
            end;

            {$IFDEF WIN32}
                ShellExecute(Application.Handle, nil, PChar(ContextString), nil, nil, SW_SHOWNORMAL)
            {$ENDIF}
               { I am sorry, you have yet to show me how to call the default browser or emailer
                from a Win16 environment }
          end
      end
  else
    inherited Notification(AComponent, Operation)
end;

procedure TURLHighlighter.setWebURL(h: THighlightInfo);
begin
  fWebURL.Assign(h)
end;

procedure TURLHighlighter.setEmailURL(h: THighlightInfo);
begin
  fEmailURL.Assign(h)
end;

procedure TURLHighlighter.setNNTP_URL(h: THighlightInfo);
begin
  fNNTP_URL.Assign(h)
end;

procedure TURLHighlighter.setFileURL(h: THighlightInfo);
begin
  fFileURL.Assign(h)
end;

procedure TURLHighlighter.setOtherURL(h: THighlightInfo);
begin
  fOtherURL.Assign(h)
end;

procedure TURLHighlighter.setScope(sc: Integer);
begin
  if sc<>fScope then
    begin
      fScope:= sc;
      HighlightChange(nil)
    end
end;

procedure TURLHighlighter.setPriority(const Value: Boolean);
begin
  if fPriority<>Value then
    begin
      fPriority:= Value;
      HighlightChange(nil)
    end;
end;


(* Now, the dirty parts  ************************************************)

{ A few utilities: }

function DynToScope(const drec: DynInfoRec): Integer;
begin
  if drec.DynStyle and $80 = 0 then Result:= 0 else Result:= drec.Context
end;


function getlevel(const dyn: DynInfoRec): SmallInt;
begin   { returns the level of nesting }
  if dyn.DynStyle and $c0 <>$c0 then Result:= -1
  else Result:= dyn.Level
end;


procedure TURLHighlighter.ApplyKeywordsList(Start, Stop: TPlusNavigator; BaseIndex: Integer); { they must be in the same par. }
  function urlok(anav: TPlusNavigator; utype: TUrlType): Boolean;
    var curscope: Integer;
    begin
      Result:= False;
      if getlevel(anav.pDynAttr^)>=15 then Exit;
      curscope:= DynToScope(anav.pDynAttr^);
      if Assigned(OnUrl) then Result:= OnUrl(Self, curscope, utype)
                         else Result:= Priority or (curscope=fScope)
    end;


var sKey, skey1     : TPlusNavigator;   { for navigating in the text when highlighting a url }
    slen            : Byte;             { length of key currently under test }
    t, tsearch      : PChar;
    tlen            : PlusCardinal;
    startparoffset  : PlusCardinal;     { real offset value of Start }

    backstart, backend, eoff: Integer;  { how much text we must examine to the left and right of Start-Stop }

    i             : PlusCardinal;  { index in t }
    skfound       : SmallInt;      { index in URLs array when a url is found }
    j             : SmallInt;      { index for looping in URLs array }
    tcomp1, tcomp2: PChar;         { used when comparing retrieved text against the keys }
    delfound      : Boolean;       { used in checking/finding the boundaries of a url just found }
    dinfo         : DynInfoRec;    { filled and passed to SetDynStyleP when a url is found }
    h: THighlightInfo;             { instance used to highlight the url just found }
    navs: Integer;                 { for iterating over TPlusNavigators if some url highlighted }
    thispar: Longint;
    smemo  : TProtPlusMemo;
    doneurl, someurl: Boolean;

begin
  smemo:= TProtPlusMemo(Start.fPMemo);
  sKey:= smemo.fssnav1;
  sKey1:= smemo.fssnav2;
  sKey.Assign(Start);
  sKey.RightOfDyn;

  thispar:= Start.ParNumber;
  if Start.fPar^.Length=0 then Exit;

  if Start.fOffset>=LongestKey+1 then backstart:= LongestKey+1
                                 else backstart:= Start.fOffset;
  skey.Pos:= Start.Pos-backstart;

  eoff:= Stop.ParOffset;
  if eoff+LongestKey+1>Stop.fPar^.Length then backend:= Stop.fPar^.Length-eoff
                                         else backend:= LongestKey+1;

  tlen:= eoff-Start.ParOffset+backstart+backend;
  t:= smemo.GetUpText(Start.fPar, thispar, skey.ParOffset, tlen);

  startparoffset:= skey.ParOffset;
  someurl:= False;

  i:= 0;
  while i<=tlen-backend do
    begin
    skfound:= -1;

    for j:= 0 to NbURLs-1 do
      with URLs[j] do
        begin
          slen:= Length(Key);
          if (i+slen+1>backstart) and (i+slen<=tlen) then
            begin
              tcomp1:= PChar(Key);
              tsearch:=t;

              {if CompareMem(@key[1], @tsearch[i], slen) then} {//lets provide a faster way!}
              tcomp2:= tsearch+i;
              while (slen>0) and (tcomp1^=tcomp2^) do
                begin
                  Dec(slen);
                  Inc(tcomp1);
                  Inc(tcomp2)
                end;
              if slen=0 then
                begin
                  slen:= Length(Key);
                  if ((StartURL and ((i=0) or (tsearch[i-1] in URLDelimiters)) or
                      (not StartURL and (i>0) and (not (tsearch[i-1] in URLDelimiters)))) and
                     ((i+slen<tlen) and (not(tsearch[i+slen] in URLDelimiters)))) then
                    begin
                      skfound:= j;
                      Break
                    end
                end  { slen=0, i.e. found URL key }
            end;  { i is in range for this key }
          end;   { j loop in URLs }

    if {not found} skfound=-1 then Inc(i)
    else
      begin  { found url key }
        doneurl:= False;
        SKey.Pos:= SKey.fPar^.StartOffset+i+startparoffset;

        if not URLs[skfound].StartURL then
          begin     { reach the start of the URL with skey}
            delfound:= False;
            while (SKey.ParOffset>0) and (not delfound) do
              begin
                SKey.Pos:= sKey.Pos-1;
                delfound:= SKey.Text in URLDelimiters
              end;
            if delfound then SKey.Pos:= SKey.Pos + 1
          end;
        skey.RightOfDyn;
        { reach the end of the url with skey1 }
        skey1.Assign(skey);
        skey1.Pos:= skey1.Pos + slen;
        dinfo.Level:= getlevel(skey1.DynAttr)+1;
        while (skey1.fOffset<skey1.fPar.Length) and (not (skey1.Text in URLDelimiters)) do
             skey1.Pos:= skey1.Pos +1;
        if (Priority or (skey1.DynNb=skey.DynNb)) and urlok(skey, URLs[skfound].URLSpecie) then
          begin
            { everything is in place: fill the dinfo fields }
            case URLs[skfound].URLSpecie of
                urlWeb: h:= fWebURL;
                urlEmail: h:= fEmailURL;
                urlNNTP:  h:= fNNTP_URL;
                urlFile:  h:= fFileURL;
                urlOther: h:= fOtherURL
                else h:= fOtherURL
              end;

            with dinfo do
              begin
                KeyIndex:= skey1.DynAttr.KeyIndex;
                DynStyle:= Byte(h.Style) or $c0;
                KeyIndex[Level]:= skfound+BaseIndex;
                Cursor:= h.Cursor;
                Backgnd:= h.Background;
                Foregnd:= h.Foreground;
                Context:= BaseContext + skfound;
                StartKlen:= skey1.Pos-skey.Pos;
                StopKLen:= StartKLen;
              end;

            { that's it }
            smemo.SetDynStyleP(skey, skey1, dinfo, True, True);

            { prepare to continue the loop }
            i:= skey1.ParOffset + startparoffset;
            doneurl:= True
          end; { urlok }

        someurl:= someurl or doneurl;
        if not doneurl then Inc(i)
      end { found url }

    end; { iloop in text buffer }

  if someurl then
      for navs:= 0 to smemo.fNavigators.Count-1 do
        with TPlusNavigator(smemo.fNavigators[navs]) do
          if (fPar<>nil) and (fParNb=thispar) then
            begin
              fDynNb:= High(PlusCardinal);
              fParLine:= High(PlusCardinal)
            end;

end;  { ApplyKeywordsList }


procedure Register;
begin
  RegisterComponents('ECM', [TURLHighlighter])
end;

var i, klen: Integer;

initialization
  for i:= 0 to NbURLs-1 do
    begin
      klen:= Length(URLs[i].Key);
      if klen>LongestKey then LongestKey:= klen
    end;

end.

